#!/usr/bin/python

"""
    ***** BEGIN LICENSE BLOCK *****
    
    Copyright (c) 2006  Center for History and New Media
                        George Mason University, Fairfax, Virginia, USA
                        http://chnm.gmu.edu
    
    Licensed under the Educational Community License, Version 1.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
    http://www.opensource.org/licenses/ecl1.php
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    ***** END LICENSE BLOCK *****
"""

API_VERSION = 2
VERSION = "2.0a2"

# Namespaces
ENV_NS = "http://schemas.xmlsoap.org/soap/envelope/"
Z_NS = "http://www.zotero.org/namespaces/SOAP"

# Field code for an item
ITEM_CODE = "ITEM"
# Field code for a new (temporary) item
NEW_ITEM_CODE = "ITEM TEMP"
# Field code for a bibliography
BIBLIOGRAPHY_CODE = "BIBL"
# Placeholder for an empty bibliography
BIBLIOGRAPHY_PLACEHOLDER = "{Bibliography}"

import xml.dom.minidom, xml.sax, xml.sax.handler, httplib, socket, sys, copy, re, platform, os

class Zotero:
	def __init__(self, doc):
		self.doc = doc
	
	def addCitation(self):
		"""Adds a citation to the current document."""
		dataString = self.doc.getDocumentData()
		if not dataString:
			# Set doc prefs if no data string yet
			self.setDocPrefs()
			dataString = self.doc.getDocumentData()
			if not dataString:
				return
		data = DocumentData(dataString)
		
		noteType = -1
		if data.style.styleClass == "note":
			noteType = data.prefs['useEndnotes']
		
		field = self.__addField(data, noteType)
		if not field:
			return
		field.setCode(NEW_ITEM_CODE)
		
		# Send request
		req = Request(self.doc, data)
		req.updateCitations(field)
		req.updateBibliography()
		req.send()
	
	def editCitation(self):
		"""Edits the citation at the cursor position."""
		dataString = self.doc.getDocumentData()
		if not dataString:
			self.doc.displayAlert("You must insert a citation before performing this operation.", 1)
			return
		data = DocumentData(dataString)
		
		field = self.doc.cursorInField(data.prefs['fieldType'])
		if not field:
			self.doc.displayAlert("You must place the cursor in a Zotero citation to edit it.", 1)
			return
		
		# Send request
		req = Request(self.doc, data)
		req.updateCitations(field)
		req.updateBibliography()
		req.send()
	
	def addBibliography(self):
		"""Adds a bibliography to the current document."""
		haveCitation = False
		dataString = self.doc.getDocumentData()
		if dataString:
			data = DocumentData(dataString)
			
			# Make sure we can have a bibliography
			if not data.style.hasBibliography:
				self.doc.displayAlert("The current bibliographic style does not define a bibliography. If you wish to add a bibliography, please choose another style.", 1)
				return
			
			# Make sure we have some citations
			for field in self.doc.getFields(data.prefs['fieldType']):
				if field.getCode().startswith(ITEM_CODE):
					haveCitation = True
		
		if not haveCitation:
			self.doc.displayAlert("You must insert a citation before performing this operation.", 1)
			return
		
		field = self.__addField(data)
		if not field:
			return
		field.setCode(BIBLIOGRAPHY_CODE)
		
		# Send request
		req = Request(self.doc, data)
		req.updateBibliography(False, True)
		req.send()
	
	def editBibliography(self):
		"""Edits bibliography metadata."""
		# Make sure we have a bibliography
		haveBibliography = False
		dataString = self.doc.getDocumentData()
		
		if dataString:
			data = DocumentData(dataString)
			
			fields = self.doc.getFields(data.prefs['fieldType'])
			
			# Iterate backwards through fields to find bibliography
			fieldList = range(len(fields))
			fieldList.reverse()
			for i in fieldList:
				if fields[i].getCode().startswith(BIBLIOGRAPHY_CODE):
					haveBibliography = True
					break
		
		if not haveBibliography:
			self.doc.displayAlert("You must insert a bibliography before performing this operation.", 1)
			return
		
		# Send request
		req = Request(self.doc, data, fields)
		req.updateBibliography(True)
		req.send()
	
	def refresh(self):
		"""Updates the citation data for all citations and bibliography entries."""
		# Make sure we have some citations
		fields = []
		dataString = self.doc.getDocumentData()
		if dataString:
			data = DocumentData(dataString)
			fields = self.doc.getFields(data.prefs['fieldType'])
		
		if not fields:
			self.doc.displayAlert("You must insert a citation before performing this operation.", 1)
			return
		
		# Send request, forcing update of citations and bibliography
		req = Request(self.doc, data)
		req.updateCitations(None, True)
		req.updateBibliography(False, True)
		req.send()
	
	def removeCodes(self):
		"""Deletes field codes."""
		fields = []
		dataString = self.doc.getDocumentData()
		if dataString:
			data = DocumentData(dataString)
			fields = self.doc.getFields(data.prefs['fieldType'])
		
		if not fields:
			self.doc.displayAlert("You must insert a citation before performing this operation.", 1)
			return
		
		result = self.doc.displayAlert("Removing field codes will prevent Zotero from updating citations and bibliographies in this document. Are you sure you want to continue?", 2, 1)
		if result:
			fields.reverse()
			for field in fields:
				field.removeCode()
	
	def setDocPrefs(self):
		"""Displays a dialog to set document preferences (style, footnotes/endnotes, etc."""
		# Get data only if there is some
		data = None
		dataString = self.doc.getDocumentData()
		if dataString:
			data = DocumentData(dataString)
		
		# Send request
		req = Request(self.doc, data)
		req.setDocPrefs()
		if data:
			req.updateCitations()
			req.updateBibliography()
		req.send()
	
	def cleanup(self):
		"""Cleans up any changes made before returning, even if an error occurred"""
		self.doc.cleanup()
	
	def __addField(self, data, noteType=-1):
		"""Checks that it is appropriate to add fields to the current document at the current
		positon, then adds one."""
		# Get citation types if necessary
		if not self.doc.canInsertField(data.prefs['fieldType']):
			self.doc.displayAlert("Zotero fields cannot be inserted here.", 1)
			return False
		
		field = self.doc.cursorInField(data.prefs['fieldType'])
		if field:
			result = self.doc.displayAlert("Replace this Zotero field?", 1, 1)
			if not result:
				return False
		
		if not field:
			field = self.doc.insertField(data.prefs['fieldType'], noteType)
		
		return field

class Request:
	editField = None
	editFieldIndex = None
	updateCitationsElement = None
	updateBibliographyElement = None
	citationsHeaderElement = None
	bibliographyHeaderElement = None
	windowAppeared = False
	halted = False
	def __init__(self, doc, data, docFields=None):
		"""Constructor for this class. Takes a Document object and a DocumentData object."""
		self.doc = doc
		self.data = data
		
		# Take docFields in constructor to minimize number of getFields calls
		self.docFields = docFields
		
		self.isMac = (platform.system() == "Darwin" and os.path.exists("/usr/bin/osascript"))
		self.xml = xml.dom.minidom.Document()
		
		# Construct SOAP request
		envelope = self.xml.createElementNS(ENV_NS, "env:Envelope")
		envelope.setAttributeNS(xml.dom.XMLNS_NAMESPACE, "xmlns:env", ENV_NS)
		envelope.setAttributeNS(xml.dom.XMLNS_NAMESPACE, "xmlns", Z_NS)
		self.xml.appendChild(envelope)
		
		self.header = self.xml.createElementNS(ENV_NS, "env:Header")
		envelope.appendChild(self.header)
		
		self.body = self.xml.createElementNS(ENV_NS, "env:Body")
		envelope.appendChild(self.body)
	
	def resetRequest(self, clearFieldHeaders=False):
		"""Remove request elements that would have been satisfied by processing the headers of the
		previous response, and optionally clear field headers from the previous request so that
		they may be regenerated"""
		for item in self.body.getElementsByTagNameNS(Z_NS, "setDocPrefs"):
			item.parentNode.removeChild(item)
		if clearFieldHeaders:
			self.docFields = None
			self.citationsHeaderElement = None
	
	def updateCitations(self, edit=None, force=False):
		"""Modifies request to update citations"""
		self.updateCitationsElement = self.xml.createElementNS(Z_NS, "updateCitations")
		
		if edit:
			self.editField = edit
			self.windowAppeared = True
		if force:
			self.updateCitationsElement.setAttribute("force", "1")
		
		self.body.appendChild(self.updateCitationsElement)
	
	def updateBibliography(self, edit=False, force=False):
		"""Modifies request to update bibliography"""
		self.updateBibliographyElement = self.xml.createElementNS(Z_NS, "updateBibliography")
		
		if edit:
			self.updateBibliographyElement.setAttribute("edit", "1")
			self.windowAppeared = True
		if force:
			self.updateBibliographyElement.setAttribute("force", "1")
		
		self.body.appendChild(self.updateBibliographyElement)
	
	def reselectItem(self, id):
		"""Reselects an item for which there is no longer an entry in the Zotero database"""
		# delete old reselectItem notifiers
		for item in self.body.getElementsByTagNameNS(Z_NS, "reselectItem"):
			item.parentNode.removeChild(item)
		
		reselectItem = self.xml.createElementNS(Z_NS, "reselectItem")
		reselectItem.setAttribute("id", id)
		self.body.appendChild(reselectItem)
		
		self.windowAppeared = True
	
	def setDocPrefs(self):
		"""Modifies request to open a dialog to set document preferences"""
		setDocPrefsElement = self.xml.createElementNS(Z_NS, "setDocPrefs")
		self.body.appendChild(setDocPrefsElement)
		
		fieldTypes = self.xml.createElementNS(Z_NS, "fieldTypes")
		for fieldType in self.doc.FIELD_SUPPORT:
			fieldTypeXML = self.xml.createElementNS(Z_NS, "fieldType")
			fieldTypeXML.setAttribute("name", fieldType)
			fieldTypes.appendChild(fieldTypeXML)
		setDocPrefsElement.appendChild(fieldTypes)
		
		self.windowAppeared = True
	
	def send(self):
		"""Sends the request itself"""
		# Generate header
		while self.header.hasChildNodes():
			self.header.removeChild(self.header.firstChild)
		
		# Add client information
		client = self.xml.createElementNS(Z_NS, "client")
		client.setAttribute("api", repr(API_VERSION))
		client.setAttribute("version", VERSION)
		client.setAttribute("moduleVersion", self.doc.VERSION)
		client.setAttribute("agent", self.doc.AGENT)
		self.header.appendChild(client)
		
		# Add document data
		if self.data:
			self.data.serializeXML(self.header)
		
		# Add field data
		if self.data:
			if not self.citationsHeaderElement:
				self.__getFields()
			self.header.appendChild(self.citationsHeaderElement)
			
			if self.bibliographyHeaderElement:
				self.header.appendChild(self.bibliographyHeaderElement)
		
		# Allow halt after reading field codes
		if self.halted:
			return
		
		requestXML = self.xml.toprettyxml(encoding="utf-8")
		response = Response(self)
		
		conn = httplib.HTTPConnection('localhost', 50001)
		try:
			conn.request("POST", "/", requestXML)
		except socket.error:
			response.parse(False, "SocketError")
			return
		except:
			response.parse(False, False)
			return
		
		if self.windowAppeared:
			if self.isMac:
				(stdin, stdout) = os.popen2('/usr/bin/osascript')
				stdout.close()
				stdin.write('tell application "System Events" to set frontmost of first process whose bundle identifier is "org.mozilla.firefox" to true')
				stdin.close()
		
		try:
			httpResponse = conn.getresponse()
		except httplib.BadStatusLine:
			response.parse(False, "BadResponse")
			return
		except:
			response.parse(False, False)
			return
		
		response.parse(httpResponse, False)
	def __getFields(self):
		"""Gets citation indices and field codes and adds them to SOAP request"""
		# If there's no document data, we definitely have no citations
		if not self.data:
			return
		
		haveBibliography = False
		
		self.citationsHeaderElement = self.xml.createElementNS(Z_NS, "citations")
		
		bibliographyFieldCode = None
		self.citationFields = []
		self.bibliographyFields = []
		
		# Get doc fields if they haven't already been gotten
		if not self.docFields:
			self.docFields = self.doc.getFields(self.data.prefs['fieldType'])
		fields = self.docFields
		
		updateWarningDisplayed = False
		for i in range(len(fields)):
			fieldCode = fields[i].getCode()
			
			# Collect only item and bibliography fields
			if fieldCode.startswith(ITEM_CODE):				# Citation field
				# Update warning hack for Zotero 1.5
				if not updateWarningDisplayed and fieldCode.find('"itemID":') != -1:
					self.halted = not self.doc.displayAlert('In order to work with Zotero 1.5 and later, your document must be permanently upgraded, and you will no longer be able to use it with Zotero 1.0. Proceed?', 2, 1)
					if self.halted:
						return
					updateWarningDisplayed = True
				
				# Add citation to XML
				citation = self.xml.createElementNS(Z_NS, "citation")
				citation.setAttribute("index", repr(len(self.citationFields)))
				if fieldCode == NEW_ITEM_CODE:
					citation.setAttribute("new", "1")
				else:
					code = self.xml.createTextNode(fieldCode[5:])
					citation.appendChild(code) 
				self.citationsHeaderElement.appendChild(citation)
				
				# Check to see if this matches a pre-specified field to be edited
				if self.editField and fields[i] == self.editField:
					self.editFieldIndex = len(self.citationFields)
					self.updateCitationsElement.setAttribute("edit", repr(self.editFieldIndex))
				
				# Add to list of citation fields
				self.citationFields.append(fields[i])
			elif fieldCode.startswith(BIBLIOGRAPHY_CODE):	# Bibliography field
				# Add to list of bibliography fields so we can update it later
				self.bibliographyFields.append(fields[i])
				bibliographyFieldCode = fieldCode
		
		if self.updateBibliographyElement and bibliographyFieldCode:
			# Check if there is additional data encoded in the bibliography
			# (e.g., uncited items)
			if bibliographyFieldCode != BIBLIOGRAPHY_CODE:
				# If there is, append it to the XML
				self.bibliographyNode = self.xml.createElementNS(Z_NS, "bibliography")
				textNode = self.xml.createTextNode(bibliographyFieldCode[len(BIBLIOGRAPHY_CODE)+1:].encode("utf-8"))
				self.bibliographyNode.appendChild(textNode)

class Response:
	def __init__(self, req):
		"""Constructor for this class. Takes a Request object."""
		self.req = req
		self.doc = req.doc
	
	def parse(self, httpResponse, error):
		"""Parses a request from the httplib Response object"""
		if self.req.windowAppeared:
			self.doc.activate()
		
		gotError = False
		if not httpResponse or not httpResponse.status:
			if error == "BadResponse":
				self.doc.displayAlert("The integration plug-in received an invalid response while communicating with Zotero. Please ensure that Zotero 1.5 or later is installed and try again.", 0)	
			elif error == "SocketError":
				self.doc.displayAlert("An error occurred contacting Zotero. Please ensure Firefox is open and try again.", 0)
			else:
				self.doc.displayAlert('An unknown error occurred contacting Zotero. Please use the "Report Errors" feature from within Zotero to report this to the Zotero developers.', 0)
			gotError = True
		elif not httpResponse.status == 200:
			# Print SOAP error
			self.xml = xml.dom.minidom.parseString(httpResponse.read())
			error = self.xml.getElementsByTagNameNS(ENV_NS, "Text")[0].firstChild.nodeValue
			self.doc.displayAlert(error.replace("$INTEGRATION_VERSION", VERSION), 0)
			gotError = True
		
		if gotError:
			# Delete new field
			if self.req.editField and self.req.editField.getCode() == NEW_ITEM_CODE:
				self.req.editField.delete()
		else:
			readXML = httpResponse.read()
			self._parseXML(readXML)
	
	def _parseXML(self, xmlString):
		"""Parses XML from a string"""
		xml.sax.parseString(xmlString, SAXHandler(self.req))

class SAXHandler(xml.sax.handler.ContentHandler):
	prefixes = {}
	defaultNamespace = ''
	
	halted = False
	callAgain = False
	soapElement = None
	textContent = ''
	
	data = None
	newData = None
	currentCitation = None
	currentCitationIndex = None
	
	def __init__(self, req):
		self.zStack = []
		self.deleteFieldIndices = []
		self.missingItems = {}
		
		self.req = req
		self.doc = req.doc
		if req.data:
			self.data = req.data
			self.citationFields = req.citationFields
			self.bibliographyFields = req.bibliographyFields
	
	def characters(self, content):
		"""Called when text content is available; we just put this into a variable"""
		if self.halted:
			return
		
		self.textContent += content
	
	def startElement(self, name, attrs):
		"""Called when a tag is started"""
		if self.halted:
			return
		
		# Go throught attrs to handle namespaces
		attrNames = attrs.keys()
		for attrName in attrNames:
			if attrName.startswith('xmlns'):
				if attrName == 'xmlns':
					self.defaultNamespace = attrs.get(attrName)
				else:
					self.prefixes[attrName[6:]] = attrs.get(attrName)
		
		namespaceURI, localName = self.__resolveNS(name)
		
		# Handle header
		if namespaceURI == ENV_NS:
			if localName != 'Envelope':
				self.soapElement = localName
		elif namespaceURI == Z_NS:
			if self.soapElement == "Header":
				if localName == "session":
					# update session ID only if necessary
					newSessionID = attrs.get("id")
					
					if not self.data:
						self.newData = DocumentData()
					elif self.data.sessionID != newSessionID:
						self.newData = copy.deepcopy(self.data)
					
					if self.newData:
						self.newData.sessionID = newSessionID
				else:
					# if any non-session ID elements are present, then we will definitely have to
					# update the document
					if not self.newData:
						self.newData = copy.deepcopy(self.data)
					
					if localName == "style":
						self.newData.style = Style(attrs.get('id'), attrs.get('class'),
							int(attrs.get('hasBibliography')))
					elif localName == "prefs":
						self.newData.prefs = {}
					elif self.zStack == ["prefs"] and localName == "pref":
						self.newData.prefs[attrs.get('name')] = int(attrs.get('value'))
			
			elif self.soapElement == "Body":
				if self.zStack == ["updateCitationsResponse", "citations"] \
						and localName == "citation":
					# keep track of the citation currently being modified
					index = int(attrs.get("index"))
					self.currentCitation = self.citationFields[index]
					self.currentCitationIndex = index
					if attrs.has_key("delete"):
						self.deleteFieldIndices.append(index)
				elif self.zStack == ["updateCitationsResponse", "citations", "citation"] \
						and localName == "missing":
					id = attrs.get("id")
					if not self.missingItems.has_key(id):
						# Missing item in citation
						if attrs.has_key("n"):
							msg = 'Item %s in this citation no longer exists in your Zotero database. Do you want to select a substitute item?' % attrs.get("n")
						else:
							msg = 'This item no longer exists in your Zotero database. Do you want to select a substitute item?'
						
						msg += '\n\n' + 'Clicking "No" will delete all citations of this item from your document.'
						self.currentCitation.select()
						self.missingItems[id] = self.doc.displayAlert(msg, 1, 3)
					
					if self.missingItems[id] == 0:		# Cancel
						self.halted = True
					elif self.missingItems[id] == 1:	# No (delete citations)
						self.deleteFieldIndices.append(self.currentCitationIndex)
						self.callAgain = True
					elif self.missingItems[id] == 2:	# Yes (reselect item)
						# Not yet implemented
						self.halted = True
						if self.newData:
							self.req.data = self.newData
						self.req.resetRequest()
						self.req.reselectItem(id)
						self.req.send()
					
			self.zStack.append(name)
	
	def endElement(self, name):
		"""Called when a tag is closed"""
		if self.halted:
			return
		
		namespaceURI, localName = self.__resolveNS(name)
		
		if namespaceURI == ENV_NS:
			if localName == "Header" and self.newData:
				# when we're done with the header, serialize new document data if necessary
				if self.data:
					self.__updateDocumentData(self.data, self.newData)
				self.doc.setDocumentData(self.newData.serialize())
		elif namespaceURI == Z_NS:
			if localName == "text" or localName == "code":
				# remove leading and trailing whitespace
				self.textContent = self.textContent.strip()
			
			if self.zStack == ["updateBibliographyResponse", "text"]:
				if self.textContent:
					for field in self.bibliographyFields:
						field.setText(self.textContent, True)
				else:
					for field in self.bibliographyFields:
						field.setText(BIBLIOGRAPHY_PLACEHOLDER, False)
			elif self.zStack == ["updateBibliographyResponse", "code"]:
				for field in self.bibliographyFields:
					field.setCode(BIBLIOGRAPHY_CODE+" "+self.textContent)
			elif self.zStack == ["updateCitationsResponse", "citations", "citation", "text"]:
				if self.textContent.find("\\", 5, -1) == -1:
					# Don't use rich text if there is
					# no backslash
					self.currentCitation.setText(self.textContent[6:-1], False)
				else:
					self.currentCitation.setText(self.textContent, True)
			elif self.zStack == ["updateCitationsResponse", "citations", "citation", "code"]:
				self.currentCitation.setCode(ITEM_CODE+" "+self.textContent)
			elif localName == "updateCitationsResponse":
				# Delete fields now, in reverse, so that we don't screw up indices
				self.deleteFieldIndices.sort()
				self.deleteFieldIndices.reverse()
				for index in self.deleteFieldIndices:
					self.citationFields[index].delete()
			
			self.zStack.pop()
		
		self.textContent = ''
	
	def endDocument(self):
		"""Called when the document has been fully parsed"""
		# Call again if requested
		if self.callAgain:
			if self.newData:
				self.req.data = self.newData
			self.req.resetRequest(True)
			self.req.send()
	
	def __resolveNS(self, name):
		"""Resolve namespaces"""
		spt = name.split(':')
		if len(spt) == 1:
			localName = name
			namespaceURI = self.defaultNamespace
		else:
			localName = spt[1]
			namespaceURI = self.prefixes[spt[0]]
		return namespaceURI, localName
	
	def __updateDocumentData(self, oldData, newData):
		"""Updates document to conform to new documentData."""
		# Only go through the following if there was old data
		fieldTypeChange = oldData and oldData.prefs['fieldType'] != newData.prefs['fieldType']
		fieldChange = oldData \
			and ((oldData.style.styleClass == newData.style.styleClass == "note" \
			and oldData.prefs['useEndnotes'] != newData.prefs['useEndnotes']) \
			or oldData.style.styleClass != newData.style.styleClass or fieldTypeChange)
		
		noteType = -1
		if newData.style.styleClass == 'note':
			noteType = newData.prefs['useEndnotes']
		
		if fieldChange:
			fields = self.citationFields
			noteTypes = [noteType]*len(self.citationFields)
			
			# Need to convert bibliography too if field type was changed
			if fieldTypeChange:
				fields.extend(self.bibliographyFields)
				noteTypes.extend([-1]*len(self.bibliographyFields))
			
			self.doc.convert(fields, newData.prefs['fieldType'], noteTypes)
			self.__refreshFields(self.doc.getFields(newData.prefs['fieldType']))
		
		# If changing to a style with no bibliography, delete bibliography
		if oldData and self.bibliographyFields and oldData.style.hasBibliography \
				and not newData.style.hasBibliography:
			deleteBib = self.doc.displayAlert("The style you have selected does not define a bibliography, but there is already an existing bibliography in your document. Would you like to remove this bibliography? If you leave the bibliography intact, it will no longer be updated.", 2, 2)
			for bibField in self.bibliographyFields:
				if deleteBib:
					bibField.delete()
				else:
					bibField.removeCode()
			# Get fields again
			self.__refreshFields(self.doc.getFields(newData.prefs['fieldType']))
		
		# In any case, save prefs
		self.doc.setDocumentData(newData.serialize())
	
	def __refreshFields(self, fields):
		"""Refreshes references to citation and bibliography fields"""
		self.bibliographyFields = []
		self.citationFields = []
		for field in fields:
			fieldCode = field.getCode()
			if fieldCode.startswith(ITEM_CODE):				# Citation field
				self.citationFields.append(field)
			elif fieldCode.startswith(BIBLIOGRAPHY_CODE):	# Bibliography field
				self.bibliographyFields.append(field)
			
class Style:
	def __init__(self, styleID="", styleClass="", hasBibliography="0"):
		self.styleID = styleID
		self.styleClass = styleClass
		self.hasBibliography = hasBibliography

class DocumentData:
	def __init__(self, string=None):
		if string:
			self.unserialize(string)
	
	def serializeXML(self, intoNode):
		"""Serializes data into XML"""
		if self.sessionID:
			doc = intoNode.ownerDocument
			
			session = doc.createElementNS(Z_NS, "session")
			session.setAttribute("id", self.sessionID)
			intoNode.appendChild(session)
			
			style = doc.createElementNS(Z_NS, "style")
			style.setAttribute("id", self.style.styleID)
			style.setAttribute("class", self.style.styleClass)
			style.setAttribute("hasBibliography", repr(self.style.hasBibliography))
			intoNode.appendChild(style)
			
			prefs = doc.createElementNS(Z_NS, "prefs")
			for key in self.prefs:
				pref = doc.createElementNS(Z_NS, "pref")
				pref.setAttribute("name", key)
				pref.setAttribute("value", repr(self.prefs[key]))
				prefs.appendChild(pref)
			intoNode.appendChild(prefs)
	
	def serialize(self):
		"""Serializes data into string form"""
		
		# Could use this code to store as XML
		"""# Generate a new XML doc
		intoDoc = xml.dom.minidom.Document()
		intoNode = intoDoc.createElementNS(Z_NS, "data")
		intoDoc.appendChild(intoNode)
		
		# Serialize into a node
		self.serializeXML(intoNode)
		
		string = intoDoc.toxml()
		
		# Get rid of the container node
		return string.replace("<data>", "", 1).replace("</data>", "", 1).replace('<?xml version="1.0" ?>', "", 1)"""
		
		# For now, store as colon-delimited for compatibility
		prefs = [self.sessionID, self.style.styleID, self.style.styleClass,
			self.style.hasBibliography, self.prefs['useEndnotes'],
			self.prefs['fieldType']]
		prefString = ""
		for pref in prefs:
			prefString += ":"+str(pref).replace(":", "::")
		return prefString[1:]
	
	def unserializeXML(self, fromNode):
		"""Loads data from XML form"""
		for node in fromNode.childNodes:
			if node.namespaceURI != Z_NS:
				continue
		
			if node.localName == "session":
				self.sessionID = node.getAttribute("id")
			elif node.localName == "style":
				self.style = Style(node.getAttribute("id"), node.getAttribute("class"),
					int(node.getAttribute("hasBibliography")))
			elif node.localName == "prefs":
				self.prefs = {}
				for pref in node.childNodes:
					if pref.namespaceURI == Z_NS and pref.localName == "pref":
						# Convert to an integer if possible
						val = pref.getAttribute("value")
						try:
							val = int(val)
						except ValueError:
							False
						self.prefs[pref.getAttribute("name")] = val
						
	def unserialize(self, string):
		"""Loads data from string form"""		
		if string[0] == "<":
			# New-style XML
			self.unserializeXML(xml.dom.minidom.parseString('<data xmlns="'+Z_NS+'">'+string+'</data>').documentElement)
		else:
			# Old-style
			splitRe = re.compile(r'(?<!\:)\:(?!\:)')
			preferenceParameters = splitRe.split(string)
			for i in range(len(preferenceParameters)):
				preferenceParameters[i] = preferenceParameters[i].replace("::", ":")
			
			if len(preferenceParameters) != 6:
				raise DataError
			
			fieldType = 0
			if preferenceParameters[5] == "True" or preferenceParameters[5] == "1":
				fieldType = 1
			
			self.sessionID = preferenceParameters[0]
			self.style = Style(preferenceParameters[1],
				preferenceParameters[2], int(preferenceParameters[3]))
			self.prefs = {'useEndnotes':int(preferenceParameters[4]),
				'fieldType':fieldType}
#!/usr/bin/python

"""
    ***** BEGIN LICENSE BLOCK *****
    
    Copyright (c) 2006  Center for History and New Media
                        George Mason University, Fairfax, Virginia, USA
                        http://chnm.gmu.edu
    
    Licensed under the Educational Community License, Version 1.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
    http://www.opensource.org/licenses/ecl1.php
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    ***** END LICENSE BLOCK *****
"""

import random, string, copy, os, tempfile, sys, platform, traceback

import uno, unohelper
from com.sun.star.task import XJobExecutor
from com.sun.star.io import XInputStream, XSeekable
from com.sun.star.awt.MessageBoxButtons import BUTTONS_OK, BUTTONS_OK_CANCEL, BUTTONS_YES_NO, \
	BUTTONS_YES_NO_CANCEL, DEFAULT_BUTTON_OK, DEFAULT_BUTTON_YES
from com.sun.star.beans import PropertyExistException	

PREFS_PROPERTY = u'ZOTERO_PREF'
BOOKMARK_REFERENCE_PROPERTY = u"ZOTERO_BREF"

FIELD_PLACEHOLDER = '{Zotero}'
PREFIX =  u'ZOTERO_'
OLD_PREFIX = u' ADDIN ZOTERO_' # Was the prefix before this was released
BIBLIOGRAPHY_CODE = "BIBL"
# A sequence of unique properties in alphabetical order that will be set to default in setText
PROPERTIES_CHANGE_TO_DEFAULT = ("CharCaseMap", "CharEscapement", "CharEscapementHeight", "CharPosture", "CharWeight")

MAX_PROPERTY_LENGTH = 255
MAX_BOOKMARK_LENGTH = 50

REFMARK_ADD_CHARS = 10
BOOKMARK_ADD_CHARS = 12

class Document:
	AGENT = "OpenOffice"
	VERSION = "2.0a1"
	FIELD_SUPPORT = [u'Fields', u'Bookmarks']
	SUPPORT_FOOTNOTES = True
	SUPPORT_ENDNOTES = True
	
	def __init__(self, ctx):
		# TODO make sure this is correctly accesed
		self.ctx = ctx
		self.desktop = self.ctx.ServiceManager.createInstance("com.sun.star.frame.Desktop")
		self.text = self.desktop.getCurrentComponent().getText()
		
		self.properties = Properties(self)
		self.tempFile = TempFile()
		
		self.isMac = (platform.system() == "Darwin" and os.path.exists("/usr/bin/osascript"))
		
		configProvider = self.ctx.ServiceManager.createInstance(
			"com.sun.star.configuration.ConfigurationProvider")
		nodepath = uno.createUnoStruct("com.sun.star.beans.PropertyValue")
		nodepath.Name = "nodepath"
		nodepath.Value = "/org.openoffice.Setup/Product"
		settings = configProvider.createInstanceWithArguments(
			"com.sun.star.configuration.ConfigurationAccess", (nodepath,))
		
		self.ooVersion = settings.getByName("ooSetupVersion")
		self.supportsInsertingDocumentFromInputStream = (int(self.ooVersion[0]) > 2 or \
			(int(self.ooVersion[0]) == 2 and int(self.ooVersion[2]) >= 4))
		
		if self.isMac:
			# Get product name
			self.__appName = settings.getByName("ooName")
			if int(self.ooVersion[0]) == 2 and self.__appName == "OpenOffice.org":
				self.__appName = "X11"
	
	def displayAlert(self, text, icon=1, buttons=0):
		"""Displays a dialog
		icon == 0 => Stop Icon
		icon == 1 => Notice Icon
		icon == 2 => Caution Icon
		
		buttons == 0 => OK
		buttons == 1 => Cancel, OK
		buttons == 2 => No, Yes
		buttons == 2 => Cancel, No, Yes"""
	
		localContext = uno.getComponentContext()
		resolver = localContext.ServiceManager.createInstanceWithContext("com.sun.star.bridge.UnoUrlResolver", localContext)
		desktop = self.ctx.ServiceManager.createInstanceWithContext("com.sun.star.frame.Desktop", self.ctx)
	
		boxTypes = ["errorbox", "messbox", "warningbox"]
		rect = uno.createUnoStruct("com.sun.star.awt.Rectangle")
		if buttons == 1:
			ooButtons = BUTTONS_OK_CANCEL + DEFAULT_BUTTON_OK
		elif buttons == 2:
			ooButtons = BUTTONS_YES_NO + DEFAULT_BUTTON_YES
		elif buttons == 3:
			ooButtons = BUTTONS_YES_NO_CANCEL + DEFAULT_BUTTON_YES
		else:
			ooButtons = BUTTONS_OK
	
		window = desktop.getCurrentFrame().getContainerWindow()# TODO make sure this is correctly accessed
		toolkit = window.getToolkit()
		msgbox = toolkit.createMessageBox(window, rect, boxTypes[icon], ooButtons, "", str(text))
		result = msgbox.execute()
		if buttons == 2:
			if result == 3:	# No
				return 0
			return 1		# Yes
		elif buttons == 3:
			if result == 3:	# No
				return 1
		elif not buttons:
			result = 0
		
		return result

	def activate(self):
		"""Brings this document to the foreground (if necessary.)"""
		# TODO make sure this works for other OpenOffice versions
		# In Ubuntu, the OpenOffice box in the taskbar blinks orange
		self.desktop.getCurrentFrame().getContainerWindow().setFocus()
		
		# On Mac, actually activate OOo
		if self.isMac:
			(stdin, stdout) = os.popen2('/usr/bin/osascript')
			stdout.close()
			stdin.write('tell application "'+self.__appName+'" to activate')
			stdin.close()
	
	def _getSelection(self):
		"""Returns the current selection in the document (specific to this implementation)"""
		return self.desktop.getCurrentFrame().getController().getViewCursor()

	def _randomChars(self, count):
		"""Returns count random chars. Used to create unique mark names
		(specific to this implementation)"""
		
		return ''.join([random.choice(string.ascii_letters + string.digits) for i in range(count)])

	def _oooObject(self, oObject, fieldType=None):
		"""Used by getFields (specific to this implementation)"""
		if oObject.ReferenceMark and (fieldType == 0 or fieldType == None):
			return oObject.ReferenceMark
		elif oObject.Bookmark and (fieldType == 1 or fieldType == None):
			return oObject.Bookmark
		elif oObject.TextSection and (fieldType == 0 or fieldType == None):
			return oObject.TextSection

	def _createNewMark(self, fieldType, field):
		"""Specific to this implementation"""
		if fieldType == 0:
			return Field(self, field)
		elif fieldType == 1:
			return Bookmark(self, field)

	def canInsertField(self, fieldType):
		"""Determines whether a field can be inserted at the current position."""
		position = self._getSelection().getText().getImplementationName()
		return (position == "SwXBodyText" or (fieldType != 1 and position == "SwXFootnote")) \
			and position != "SwXCell" # OOo bug causes error when inside a nested table inside the main body

	def cursorInField(self, fieldType):
		"""Determines whether the cursor is in a mark.
		Returns the mark if it is, or false if it is not."""
		thisController = self.desktop.getCurrentFrame().getController()
		cursor = self._getSelection()
		oRange1 = cursor.getText().createTextCursorByRange(cursor)
		oRange2 = cursor.getText().createTextCursorByRange(cursor)
		oRange1.goLeft(1, False)
		oRange1.gotoStartOfParagraph(True)
		oRange2.gotoEndOfParagraph(True)
		oPortionEnum1 = oRange1.createEnumeration().nextElement().createEnumeration()
		while oPortionEnum1.hasMoreElements():
			oPortion1 = oPortionEnum1.nextElement()
			if self._oooObject(oPortion1, fieldType) and oPortion1.IsStart:
				sName = self._oooObject(oPortion1).getName()
				if sName.startswith(PREFIX) or sName.startswith(OLD_PREFIX):
					oPortionEnum2 = oRange2.createEnumeration().nextElement().createEnumeration()
					while oPortionEnum2.hasMoreElements():
						oPortion2 = oPortionEnum2.nextElement()
						if self._oooObject(oPortion2, fieldType) and not oPortion2.IsStart:
							if self._oooObject(oPortion2).getName() == sName:
								mark = self._oooObject(oPortion2)
								thisController.select(mark.getAnchor())
								return self._createNewMark(fieldType, mark)
		return False
	
	def getDocumentData(self):
		"""Retrieves persistent data from this document. Returns a string."""
		return self.properties.getProperty(PREFS_PROPERTY)

	def setDocumentData(self, documentData):
		"""Sets persistent data for this document. documentData is a string."""
		self.properties.setProperty(PREFS_PROPERTY, documentData)

	def insertField(self, fieldType, noteType=-1, rangeToInsert=None, markName=None, bookmarkName=None):
		"""Makes a field at the selection. The "rangeToInsert" and "markName"
		arguments are specific to this implementation (both are used by convert
		below) and need not be implemented."""
		thisComponent = self.desktop.getCurrentComponent()
		if rangeToInsert == None:
			rangeToInsert = self._getSelection()
		if not markName:
			markName = ""

		# Make footnote or endnote if desired (and cursor is in main text)
		locationType = rangeToInsert.getText().getImplementationName()
		if noteType != -1 and locationType == "SwXBodyText":
			oNote = None
			if noteType == 0: # Footnote
				oNote = thisComponent.createInstance("com.sun.star.text.Footnote")
			elif noteType == 1: # Endnote
				oNote = thisComponent.createInstance("com.sun.star.text.Endnote")
			rangeToInsert.getText().insertTextContent(rangeToInsert, oNote, True)
			rangeToInsert = oNote.getText().createTextCursorByRange(oNote)
		
		rangeToInsert.setString(FIELD_PLACEHOLDER)

		# Make Field
		if fieldType == 0:		# Field
			oField = thisComponent.createInstance("com.sun.star.text.ReferenceMark")
			oField.setName(PREFIX + markName + "_" + self._randomChars(REFMARK_ADD_CHARS))
		elif fieldType == 1:	# Bookmark
			oField = thisComponent.createInstance("com.sun.star.text.Bookmark")
			if bookmarkName == None:
				bookmarkName = BOOKMARK_REFERENCE_PROPERTY +"_"+ self._randomChars(BOOKMARK_ADD_CHARS)
			oField.setName(bookmarkName)

		oField.attach(rangeToInsert)
		mark = self._createNewMark(fieldType, oField)
		if fieldType == 1:
			mark.setCode(markName)
		return mark

	def getFields(self, fieldType):
		"""Gets a sorted list of all fields from this document."""
		fields = []
		if fieldType == 0:		# Fields
			marks = self.desktop.getCurrentComponent().getReferenceMarks()
		elif fieldType == 1:	# Bookmarks
			marks = self.desktop.getCurrentComponent().getBookmarks()

		markNames = marks.getElementNames()
		if markNames:
			for name in markNames:
				if name.startswith(PREFIX) or name.startswith(OLD_PREFIX):
					mark = marks.getByName(name)
					fields.append(self._createNewMark(fieldType, mark))

		if fieldType == 0:
			marks = self.desktop.getCurrentComponent().getTextSections() # Bibliography for Fields
			markNames = marks.getElementNames()
			if markNames:
				for name in markNames:
					if name.startswith(PREFIX) or name.startswith(OLD_PREFIX):
						mark = marks.getByName(name)
						fields.append(self._createNewMark(fieldType, mark))
		fields.sort()
		return fields

	def convert(self, fields, toFieldType, noteTypes):
		"""Convert "fields" to a different fieldType, or a different noteType."""
		newFields = []
		for field, noteType in zip(fields, noteTypes):
			fieldCode = field.getCode()
			oRange = field.delete()
			rangeToInsert = oRange.getText().createTextCursorByRange(oRange)
			newFields.append(self.insertField(toFieldType, noteType, rangeToInsert, fieldCode, field.bookmarkName))

	def cleanup(self):
		"""Run on exit to clean up anything we played with..."""
		if self.tempFile.path:
			os.unlink(self.tempFile.path)
		
		# NOTE: Not sure this is necessary with OOo > 2.0
		# From: http://udk.openoffice.org/python/python-bridge.html
		# Do a nasty thing before exiting the python process. In case the
		# last call is a oneway call (e.g. see idl-spec of insertString),
		# it must be forced out of the remote-bridge caches before python
		# exits the process. Otherwise, the oneway call may or may not reach
		# the target object.
		# I do this here by calling a cheap synchronous call (getPropertyValue).
		self.ctx.ServiceManager

class Field:
	def __init__(self, wpDoc, field):
		self.wpDoc = wpDoc
		self.fieldType = 0 # Field
		self.field = field
		self.bookmarkName = None
		if self._isBibliography():
			self.__convertToTextSection()

	def __cmp__(x, y):
		xRange = x.field.getAnchor()
		text = xRange.getText()
		if text.supportsService("com.sun.star.text.Footnote"):
			xRange = text.getAnchor()
		yRange = y.field.getAnchor()
		text = yRange.getText()
		if text.supportsService("com.sun.star.text.Footnote"):
			yRange = text.getAnchor()
		return x.wpDoc.text.compareRegionStarts(yRange, xRange)

	def _isBibliography(self):
		"""Specific to this implementation"""
		return self.getCode().startswith(BIBLIOGRAPHY_CODE)
	
	def _insertRTF(self, oRange, string):
		"""Inserts RTF string at oRange. Specific to this implementation"""
		if(self.wpDoc.supportsInsertingDocumentFromInputStream):
			filterName = uno.createUnoStruct("com.sun.star.beans.PropertyValue")
			filterName.Name = "FilterName"
			filterName.Value = "Rich Text Format"
			inputStream = uno.createUnoStruct("com.sun.star.beans.PropertyValue")
			inputStream.Name = "InputStream"
			inputStream.Value = StringInputStream(string)
			oRange.insertDocumentFromURL("private:stream", (filterName, inputStream))
		else:
			# Can't use a private input stream to insert a document if OOo < 2.4
			# See http://api.openoffice.org/issues/show_bug.cgi?id=83449
			self.wpDoc.tempFile.write(string)
			oRange.insertDocumentFromURL(uno.systemPathToFileUrl(self.wpDoc.tempFile.path), ())
			

	def __isTextSection(self):
		""" Text Sections are used for Bibliographies because Reference Marks
		 can not span more than one line (specific to this implementation)"""
		return self.field.supportsService("com.sun.star.text.TextSection")

	def __convertToTextSection(self):
		"""Converts a ReferenceMark to a TextSection for bibliographies (specific to this implementation)"""
		code = self.getFullCode()
		oRange = self.removeCode()
		self.field = self.wpDoc.desktop.getCurrentComponent().createInstance("com.sun.star.text.TextSection")
		self.field.setName(code)
		self.field.attach(oRange)
	
	def __getOutOfField(self):
		"""Gets the cursor out of the OOo ReferenceMark, if it is currently inside it"""
		try:
			cursor = self.wpDoc._getSelection()
			if cursor.isCollapsed() and self.wpDoc.text.compareRegionEnds(cursor, self.field.getAnchor()) == 0:
				cursor.gotoRange(self.field.getAnchor().getEnd(), False)
				thisController = self.wpDoc.desktop.getCurrentFrame().getController()
				self.wpDoc.ctx.ServiceManager.createInstance("com.sun.star.frame.DispatchHelper"). \
					executeDispatch(thisController, ".uno:ResetAttributes", "", 0, ())
		except:
			pass

	def delete(self):
		"""Deletes this field and returns the range
		 (return value is specific to this implementation)"""
		oRange = self.field.getAnchor()
		text = oRange.getText()
		if text.supportsService("com.sun.star.text.Footnote"):
			mainRange = text.getAnchor()
			text.dispose()
			return mainRange
		else:
			oRange.setString('')
			if self.__isTextSection():
				self.field.dispose()
			return oRange

	def removeCode(self):
		"""Removes field, but maintains its contents
		and returns the range (return value is specific to this implementation)"""
		oRange = self.field.getAnchor()
		if self.__isTextSection():
			self.field.dispose()
		else:
			oRange.setString(oRange.getString())
		return oRange

	def select(self):
		"""Selects this field"""
		cursor = self.wpDoc._getSelection()
		cursor.gotoRange(self.field.getAnchor(), False)
		if self.__isTextSection():
			cursor.goLeft(1, True)

	def setText(self, string, isRich):
		"""Sets the text inside this field to a specified pseudo-RTF formatted text string"""
		oRange = self.field.getAnchor()
		oRange = oRange.getText().createTextCursorByRange(oRange)
		oRange.setPropertiesToDefault(PROPERTIES_CHANGE_TO_DEFAULT)
		
		if not self.__isTextSection():
			code = self.getFullCode()
			
		if isRich:
			if self.__isTextSection():	# Bibliography
				# Add a new line to the start of the bibliography so that the paragraph format
				# for the first entry will be correct. Without the new line, when converting
				# citation styles, the first entry of the bibliography will keep the same paragraph
				# formatting as the previous citation style
				string = "{\\rtf\\\n" + string[6:]
			
			self._insertRTF(oRange, string)
			
			if self.__isTextSection():
				# Remove the new line from the bibliography (added above). Have to remove the
				# new line before the textSection and then adjust the range so the new line
				# starting the textSection is outside of the range so that the 
				# paragraph formatting of the first entry remains unchanged. Also remove the
				# extra new line at the end of the textSection.
				nLen = len(oRange.getString())
				removeLastNewLine = 0
				if oRange.getString()[-1] == chr(10):
					removeLastNewLine = 1
					oDupRange = oRange.getText().createTextCursorByRange(oRange)
					oDupRange.collapseToEnd()
					oDupRange.goLeft(1, True)
					oDupRange.setString("")
				oRange.collapseToStart()
				oRange.goLeft(1, True)
				oRange.setString("")
				oRange.goRight(1, False)
				oRange.goRight(nLen - 1 - removeLastNewLine, True)
		else:
			oRange.setString(string)
		
		if not self.__isTextSection():
			self.field = self.wpDoc.desktop.getCurrentComponent().createInstance("com.sun.star.text.ReferenceMark")
			self.field.setName(code)
			self.field.attach(oRange)
			
			self.__getOutOfField()
				
	def setCode(self, code):
		"""Sets some (non-user-readable) code to accompany this field"""
		if not self.__isTextSection() and code.startswith(BIBLIOGRAPHY_CODE):
			self.__convertToTextSection()
		code = PREFIX + code + " RND" + self.wpDoc._randomChars(REFMARK_ADD_CHARS)
		if self.__isTextSection():
			self.field.setName(code)
		else:
			# The only way of renaming a referencemark is to delete it and recreate it with the new name
			# TODO: This isn't going to work with formatted text
			oRange = self.field.getAnchor()
			sString = oRange.getString()
			oRange.setString(sString)
			self.field = self.wpDoc.desktop.getCurrentComponent().createInstance("com.sun.star.text.ReferenceMark")
			self.field.setName(code)
			self.field.attach(oRange)
		
	def getCode(self):
		"""Returns the current code"""
		fieldName = self.field.getName()
		
		# Find offset of random data
		rnd = fieldName.rfind(" RND")
		if rnd == -1:	# For compatiblity with old Python plug-in
			rnd = -6
		
		if fieldName.startswith(OLD_PREFIX):
			self.setCode(fieldName[len(OLD_PREFIX):rnd])
			return fieldName[len(OLD_PREFIX):rnd]
		
		return fieldName[len(PREFIX):rnd]

	def getFullCode(self):
		"""Returns the full current code"""
		return self.field.getName()

class Bookmark(Field):
	def __init__(self, wpDoc, field):
		self.wpDoc = wpDoc
		self.fieldType = 1
		self.field = field
		self.bookmarkName = field.getName()

	def delete(self):
		"""Deletes this bookmark, and any properties
		and returns the range (return value is specific to this implementation)"""
		self.wpDoc.properties.setProperty(self.bookmarkName, "")
		oRange = self.field.getAnchor()
		if oRange.getText().supportsService("com.sun.star.text.Footnote"):
			oMainRange = oRange.getText().getAnchor()
			oRange.getText().dispose()
			return oMainRange
		else:
			self.field.dispose()
			oRange.setString('')
			return oRange

	def removeCode(self):
		"""Removes bookmark, but maintains its contents
		and returns the range (return value is specific to this implementation)"""
		self.wpDoc.properties.setProperty(self.bookmarkName, "")
		oRange = self.field.getAnchor()
		self.field.dispose()
		return oRange

	def setText(self, string, isRich):
		"""Sets the text inside this field to a specified pseudo-RTF formatted
		   text string"""
		oRange = self.field.getAnchor()
		oRange = oRange.getText().createTextCursorByRange(oRange)
		oRange.setPropertiesToDefault(PROPERTIES_CHANGE_TO_DEFAULT)
		if isRich:
			if self._isBibliography():
				# Add a new line to the start of the bibliography so that the paragraph format
				# for the first entry will be correct. Without the new line, when converting
				# citation styles, the first entry of the bibliography will keep the same paragraph
				# formatting as the previous citation style
				string = "{\\rtf\\\n" + string[6:]

				# Make sure bibliography starts on its own line
				oDupRange = oRange.getText().createTextCursorByRange(oRange)
				oDupRange.collapseToStart()
				oDupRange.goLeft(1, True)
				if ord(oDupRange.getString()[-1]) != 10:
					oDupRange.collapseToEnd()
					oDupRange.setString("\n")

			self._insertRTF(oRange, string)
			
			if self._isBibliography():
				# Remove the new line from the bibliography (added above). Have to remove the
				# new line before the bookmark and then adjust the range so the new line
				# starting the bookmark is outside of the range so that the 
				# paragraph formatting of the first entry remains unchanged. Also remove the
				# extra new line at the end of the bookmark.
				sLen = len(oRange.getString())
				removeLastNewLine = 0
				if oRange.getString()[-1] == chr(10):
					removeLastNewLine = 1
					oDupRange = oRange.getText().createTextCursorByRange(oRange)
					oDupRange.collapseToEnd()
					# These two seemingly redundant lines actually make sure we don't
					# change the paragraph formatting of the last entry in the bibliography
					oDupRange.goLeft(1, False)
					oDupRange.goRight(1, True)

					oDupRange.setString("")

				oRange.collapseToStart()
				oRange.goLeft(1, True)
				oRange.setString("")
				oRange.goRight(1, False)
				oRange.goRight(sLen - 1 - removeLastNewLine, True)
			
			self.field.dispose()
			self.field = self.wpDoc.desktop.getCurrentComponent().createInstance("com.sun.star.text.Bookmark")
			self.field.setName(self.bookmarkName)
			self.field.attach(oRange)
		else:
			oRange.setString(string)

	def setCode(self, code):
		"""Sets some (non-user-readable) code to accompany this bookmark"""
		# Set the property corresponding to the name
		self.wpDoc.properties.setProperty(self.bookmarkName, PREFIX+code)

	def getCode(self):
		"""Returns the current code"""
		return self.wpDoc.properties.getProperty(self.bookmarkName)[len(PREFIX):]

	def getFullCode(self):
		"""Returns the current code
		This function is specific to this implementation"""
		return self.wpDoc.properties.getProperty(self.bookmarkName)

class Properties:
	def __init__(self, wpDoc):
		self.wpDoc = wpDoc
		self.asProperties = wpDoc.desktop.getCurrentComponent().getDocumentInfo()

	def getProperty(self, property):
		"""Gets a document-specific property."""
		i = 0
		propertyValue = ""

		while True:
			i = i + 1
			try:
				propertyValue += self.asProperties.getPropertyValue(property+"_"+`i`)
			except:
				break
		return propertyValue

	def setProperty(self, property, value):
		"""Sets a document-specific property."""
		i = 0

		# Update existing property values or add new ones
		while value != "":
			i = i + 1
			docPropertyName = property+u'_'+`i`
			docPropertyValue = value[0:MAX_PROPERTY_LENGTH]
			value = value[MAX_PROPERTY_LENGTH:]
			try:
				self.asProperties.addProperty(docPropertyName, 0, "")
			except PropertyExistException:
				pass
			self.asProperties.setPropertyValue(docPropertyName, docPropertyValue)

		# Delete unnecessary fields
		while True:
			i = i + 1
			try:
				self.asProperties.removeProperty(property+u'_'+`i`)
			except:
				break

class StringInputStream(unohelper.Base, XInputStream, XSeekable):
	"""Implementation of XInputStream; used for XML facilities and to insert RTF text into the
	document if OOo is version 2.4 or newer (see _insertRTF)"""
	def __init__(self, text):
		if isinstance(text, unicode):
			self.text = text.encode("utf8")
		else:
			self.text = text
		self.position = 0
	
	def seek(self, position):
		self.position = position
	
	def getPosition(self):
		return self.position
	
	def getLength(self):
		return len(self.text)
	
	def readBytes(self, data, bytesToRead):
		data = self.text[self.position:(self.position+bytesToRead)]
		dataLength = len(data)
		self.position += dataLength
		return dataLength, uno.ByteSequence(data)
   
	def readSomeBytes(self, data, bytesToRead):
		return self.readBytes(data, bytesToRead)

	def skipBytes(self, bytesToSkip):
		self.position += bytesToSkip;

	def available(self):
		return len(self.text) - self.position;

	def closeInput(self):
		self.text = None

class TempFile:
	"""Used to insert RTF text into the document if OOo is older than version 2.4 (see _insertRTF)"""
	path = None
	def write(self, string):
		if not self.path:
			(file, self.path) = tempfile.mkstemp('.rtf')
			file = os.fdopen(file, "w")
		file.write(string)
		file.close()

if Zotero:
	# Register UNO Component
	class ZoteroUnoComponent(unohelper.Base, XJobExecutor):
		def __init__(self, ctx):
			self.doc = Document(ctx)
			self.z = Zotero(self.doc)
			
			# On OOo 3, there seem to be some issues with pyexpat, so we use OOo's SAX support
			# instead of Python's. This is an ugly hack, and almost certainly much slower than
			# pyexpat because of the pyuno bridge overhead, but it works.
			from com.sun.star.xml.sax import XDocumentHandler
			
			class OOoSAXHandler(SAXHandler, unohelper.Base, XDocumentHandler):
				"""Implementation of XDocumentHandler"""
				def startElement(self, name, attrs):
					"""Convert attrs from OOo format to Python format"""
					newAttrs = {}
					for i in range(attrs.getLength()):
						newAttrs[attrs.getNameByIndex(i)] = attrs.getValueByIndex(i)
					SAXHandler.startElement(self, name, newAttrs)
			
			def _parseXML(self, xmlString):
				"""Substitute _parseXML function that avoids pyexpat"""
				parser = ctx.ServiceManager.createInstance("com.sun.star.xml.sax.Parser")
				handler = OOoSAXHandler(self.req)
				parser.setDocumentHandler(handler)
				inputSrc = uno.createUnoStruct("com.sun.star.xml.sax.InputSource")
				inputSrc.aInputStream = StringInputStream(xmlString)
				parser.parseStream(inputSrc) 
			
			Response._parseXML = _parseXML
		def trigger(self, command):
			try:
				try:
					getattr(self.z, command)()
				except:
					info = sys.exc_info()
					exception = traceback.format_exception(info[0], info[1], info[2])
					self.doc.displayAlert("An unexpected error occured while performing this operation:\n\n"+\
						"\n"+''.join(exception[-4:]), 0)
			finally:
				self.z.cleanup()
	
	g_ImplementationHelper = unohelper.ImplementationHelper()
	g_ImplementationHelper.addImplementation(ZoteroUnoComponent, "org.Zotero.integration.openoffice.Functions",
		("com.sun.star.task.Job",),)
else:
	# Called from command-line
	from Zotero import Zotero
	
	localContext = uno.getComponentContext()
	resolver = localContext.ServiceManager.createInstanceWithContext("com.sun.star.bridge.UnoUrlResolver", localContext)
	ctx = resolver.resolve( "uno:socket,host=localhost,port=2002;urp;StarOffice.ComponentContext")
	
	z = Zotero(Document(ctx))
	
	try:
		getattr(z, sys.argv[1])()
	finally:
		z.cleanup()